package com.sun.corba.se.impl.interceptors;

import java.util.Stack;
import org.omg.CORBA.PolicyError;
import org.omg.PortableInterceptor.PolicyFactory;
import org.omg.CORBA.Policy;
import org.omg.CORBA.BAD_PARAM;
import org.omg.PortableInterceptor.Current;
import org.omg.PortableInterceptor.ORBInitInfoPackage.DuplicateName;
import org.omg.PortableInterceptor.Interceptor;
import org.omg.PortableInterceptor.ORBInitializer;
import com.sun.corba.se.spi.ior.ObjectKeyTemplate;
import org.omg.CORBA.Any;
import org.omg.CORBA.NVList;
import com.sun.corba.se.spi.protocol.ForwardException;
import com.sun.corba.se.impl.protocol.giopmsgheaders.ReplyMessage;
import com.sun.corba.se.impl.corba.RequestImpl;
import com.sun.corba.se.spi.protocol.CorbaMessageMediator;
import org.omg.CORBA.CompletionStatus;
import com.sun.corba.se.spi.protocol.RetryType;
import org.omg.CORBA.portable.ApplicationException;
import org.omg.CORBA.UserException;
import org.omg.CORBA.portable.RemarshalException;
import org.omg.CORBA.SystemException;
import org.omg.PortableInterceptor.ObjectReferenceTemplate;
import com.sun.corba.se.spi.oa.ObjectAdapter;
import com.sun.corba.se.spi.orbutil.closure.ClosureFactory;
import java.util.HashMap;
import org.omg.IOP.CodecFactory;
import com.sun.corba.se.impl.logging.OMGSystemException;
import com.sun.corba.se.impl.logging.ORBUtilSystemException;
import com.sun.corba.se.impl.logging.InterceptorsSystemException;
import com.sun.corba.se.spi.orb.ORB;
import com.sun.corba.se.spi.protocol.PIHandler;

public class PIHandlerImpl implements PIHandler
{
    boolean printPushPopEnabled;
    int pushLevel;
    private ORB orb;
    InterceptorsSystemException wrapper;
    ORBUtilSystemException orbutilWrapper;
    OMGSystemException omgWrapper;
    private int serverRequestIdCounter;
    CodecFactory codecFactory;
    String[] arguments;
    private InterceptorList interceptorList;
    private boolean hasIORInterceptors;
    private boolean hasClientInterceptors;
    private boolean hasServerInterceptors;
    private InterceptorInvoker interceptorInvoker;
    private PICurrent current;
    private HashMap policyFactoryTable;
    private static final short[] REPLY_MESSAGE_TO_PI_REPLY_STATUS;
    private ThreadLocal threadLocalClientRequestInfoStack;
    private ThreadLocal threadLocalServerRequestInfoStack;
    
    private void printPush() {
        if (!this.printPushPopEnabled) {
            return;
        }
        this.printSpaces(this.pushLevel);
        ++this.pushLevel;
        System.out.println("PUSH");
        return;
    }
    
    private void printPop() {
        if (!this.printPushPopEnabled) {
            return;
        }
        this.printSpaces(--this.pushLevel);
        System.out.println("POP");
        return;
    }
    
    private void printSpaces(int n) {
        for (int i = 0; i < n; ++i) {
            System.out.print(" ");
        }
    }
    
    public void close() {
        this.orb = null;
        this.wrapper = null;
        this.orbutilWrapper = null;
        this.omgWrapper = null;
        this.codecFactory = null;
        this.arguments = null;
        this.interceptorList = null;
        this.interceptorInvoker = null;
        this.current = null;
        this.policyFactoryTable = null;
        this.threadLocalClientRequestInfoStack = null;
        this.threadLocalServerRequestInfoStack = null;
    }
    
    public PIHandlerImpl(ORB orb, String[] arguments) {
        this.printPushPopEnabled = false;
        this.pushLevel = 0;
        this.serverRequestIdCounter = 0;
        this.codecFactory = null;
        this.arguments = null;
        this.threadLocalClientRequestInfoStack = new ThreadLocal() {
            @Override
            protected Object initialValue() {
                return new PIHandlerImpl.RequestInfoStack();
            }
        };
        this.threadLocalServerRequestInfoStack = new ThreadLocal() {
            @Override
            protected Object initialValue() {
                return new PIHandlerImpl.RequestInfoStack();
            }
        };
        this.orb = orb;
        this.wrapper = InterceptorsSystemException.get(orb, "rpc.protocol");
        this.orbutilWrapper = ORBUtilSystemException.get(orb, "rpc.protocol");
        this.omgWrapper = OMGSystemException.get(orb, "rpc.protocol");
        this.arguments = arguments;
        this.codecFactory = new CodecFactoryImpl(orb);
        this.interceptorList = new InterceptorList(this.wrapper);
        this.current = new PICurrent(orb);
        this.interceptorInvoker = new InterceptorInvoker(orb, this.interceptorList, this.current);
        orb.getLocalResolver().register("PICurrent", ClosureFactory.makeConstant(this.current));
        orb.getLocalResolver().register("CodecFactory", ClosureFactory.makeConstant(this.codecFactory));
    }
    
    public void initialize() {
        if (this.orb.getORBData().getORBInitializers() != null) {
            ORBInitInfoImpl orbInitInfo = this.createORBInitInfo();
            this.current.setORBInitializing(true);
            this.preInitORBInitializers(orbInitInfo);
            this.postInitORBInitializers(orbInitInfo);
            this.interceptorList.sortInterceptors();
            this.current.setORBInitializing(false);
            orbInitInfo.setStage(2);
            this.hasIORInterceptors = this.interceptorList.hasInterceptorsOfType(2);
            this.hasClientInterceptors = true;
            this.hasServerInterceptors = this.interceptorList.hasInterceptorsOfType(1);
            this.interceptorInvoker.setEnabled(true);
        }
    }
    
    public void destroyInterceptors() {
        this.interceptorList.destroyAll();
    }
    
    public void objectAdapterCreated(ObjectAdapter objectAdapter) {
        if (!this.hasIORInterceptors) {
            return;
        }
        this.interceptorInvoker.objectAdapterCreated(objectAdapter);
        return;
    }
    
    public void adapterManagerStateChanged(int n, short n2) {
        if (!this.hasIORInterceptors) {
            return;
        }
        this.interceptorInvoker.adapterManagerStateChanged(n, n2);
        return;
    }
    
    public void adapterStateChanged(ObjectReferenceTemplate[] array, short n) {
        if (!this.hasIORInterceptors) {
            return;
        }
        this.interceptorInvoker.adapterStateChanged(array, n);
        return;
    }
    
    public void disableInterceptorsThisThread() {
        if (!this.hasClientInterceptors) {
            return;
        }
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get();
        ++requestInfoStack.disableCount;
        return;
    }
    
    public void enableInterceptorsThisThread() {
        if (!this.hasClientInterceptors) {
            return;
        }
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get();
        --requestInfoStack.disableCount;
        return;
    }
    
    public void invokeClientPIStartingPoint() throws RemarshalException {
        if (!this.hasClientInterceptors) {
            return;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return;
        }
        ClientRequestInfoImpl peekClientRequestInfoImplStack = this.peekClientRequestInfoImplStack();
        this.interceptorInvoker.invokeClientInterceptorStartingPoint(peekClientRequestInfoImplStack);
        int replyStatus = peekClientRequestInfoImplStack.getReplyStatus();
        if (replyStatus == 1 || replyStatus == 3) {
            Exception invokeClientPIEndingPoint = this.invokeClientPIEndingPoint(this.convertPIReplyStatusToReplyMessage((short)replyStatus), peekClientRequestInfoImplStack.getException());
            if ((invokeClientPIEndingPoint instanceof SystemException)) {
                throw (SystemException)invokeClientPIEndingPoint;
            }
            if ((invokeClientPIEndingPoint instanceof RemarshalException)) {
                throw (RemarshalException)invokeClientPIEndingPoint;
            }
            if (invokeClientPIEndingPoint instanceof UserException || invokeClientPIEndingPoint instanceof ApplicationException) {
                throw this.wrapper.exceptionInvalid();
            }
        }
        else if (replyStatus != -1) {
            throw this.wrapper.replyStatusNotInit();
        }
        return;
    }
    
    public Exception makeCompletedClientRequest(int n, Exception ex) {
        return this.handleClientPIEndingPoint(n, ex, false);
    }
    
    public Exception invokeClientPIEndingPoint(int n, Exception ex) {
        return this.handleClientPIEndingPoint(n, ex, true);
    }
    
    public Exception handleClientPIEndingPoint(int n, Exception exception, boolean b) {
        if (!this.hasClientInterceptors) {
            return exception;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return exception;
        }
        int replyStatus = PIHandlerImpl.REPLY_MESSAGE_TO_PI_REPLY_STATUS[n];
        ClientRequestInfoImpl peekClientRequestInfoImplStack = this.peekClientRequestInfoImplStack();
        peekClientRequestInfoImplStack.setReplyStatus((short)replyStatus);
        peekClientRequestInfoImplStack.setException(exception);
        if (b) {
            this.interceptorInvoker.invokeClientInterceptorEndingPoint(peekClientRequestInfoImplStack);
            replyStatus = peekClientRequestInfoImplStack.getReplyStatus();
        }
        if (replyStatus == 3 || replyStatus == 4) {
            peekClientRequestInfoImplStack.reset();
            if (b) {
                peekClientRequestInfoImplStack.setRetryRequest(RetryType.AFTER_RESPONSE);
            }
            else {
                peekClientRequestInfoImplStack.setRetryRequest(RetryType.BEFORE_RESPONSE);
            }
            exception = new RemarshalException();
        }
        else if (replyStatus == 1 || replyStatus == 2) {
            exception = peekClientRequestInfoImplStack.getException();
        }
        return exception;
    }
    
    public void initiateClientPIRequest(boolean b) {
        if (!this.hasClientInterceptors) {
            return;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return;
        }
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get();
        Object item = null;
        if (!requestInfoStack.empty()) {
            item = (ClientRequestInfoImpl)requestInfoStack.peek();
        }
        if (!b && item != null && item.isDIIInitiate()) {
            item.setDIIInitiate(false);
        }
        else {
            if (item == null || !item.getRetryRequest().isRetry()) {
                item = new ClientRequestInfoImpl(this.orb);
                requestInfoStack.push(item);
                this.printPush();
            }
            item.setRetryRequest(RetryType.NONE);
            item.incrementEntryCount();
            item.setReplyStatus((short)(-1));
            if (b) {
                item.setDIIInitiate(true);
            }
        }
        return;
    }
    
    public void cleanupClientPIRequest() {
        if (!this.hasClientInterceptors) {
            return;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return;
        }
        ClientRequestInfoImpl peekClientRequestInfoImplStack = this.peekClientRequestInfoImplStack();
        if (!peekClientRequestInfoImplStack.getRetryRequest().equals(RetryType.BEFORE_RESPONSE) && peekClientRequestInfoImplStack.getReplyStatus() == -1) {
            this.invokeClientPIEndingPoint(2, this.wrapper.unknownRequestInvoke(CompletionStatus.COMPLETED_MAYBE));
        }
        peekClientRequestInfoImplStack.decrementEntryCount();
        if (peekClientRequestInfoImplStack.getEntryCount() == 0 && !peekClientRequestInfoImplStack.getRetryRequest().isRetry()) {
            ((PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get()).pop();
            this.printPop();
        }
        return;
    }
    
    public void setClientPIInfo(CorbaMessageMediator info) {
        if (!this.hasClientInterceptors) {
            return;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return;
        }
        this.peekClientRequestInfoImplStack().setInfo(info);
        return;
    }
    
    public void setClientPIInfo(RequestImpl diiRequest) {
        if (!this.hasClientInterceptors) {
            return;
        }
        if (!this.isClientPIEnabledForThisThread()) {
            return;
        }
        this.peekClientRequestInfoImplStack().setDIIRequest(diiRequest);
        return;
    }
    
    public void invokeServerPIStartingPoint() {
        if (!this.hasServerInterceptors) {
            return;
        }
        ServerRequestInfoImpl peekServerRequestInfoImplStack = this.peekServerRequestInfoImplStack();
        this.interceptorInvoker.invokeServerInterceptorStartingPoint(peekServerRequestInfoImplStack);
        this.serverPIHandleExceptions(peekServerRequestInfoImplStack);
        return;
    }
    
    public void invokeServerPIIntermediatePoint() {
        if (!this.hasServerInterceptors) {
            return;
        }
        ServerRequestInfoImpl peekServerRequestInfoImplStack = this.peekServerRequestInfoImplStack();
        this.interceptorInvoker.invokeServerInterceptorIntermediatePoint(peekServerRequestInfoImplStack);
        peekServerRequestInfoImplStack.releaseServant();
        this.serverPIHandleExceptions(peekServerRequestInfoImplStack);
        return;
    }
    
    public void invokeServerPIEndingPoint(ReplyMessage replyMessage) {
        if (!this.hasServerInterceptors) {
            return;
        }
        ServerRequestInfoImpl peekServerRequestInfoImplStack = this.peekServerRequestInfoImplStack();
        peekServerRequestInfoImplStack.setReplyMessage(replyMessage);
        peekServerRequestInfoImplStack.setCurrentExecutionPoint(2);
        if (!peekServerRequestInfoImplStack.getAlreadyExecuted()) {
            int replyStatus = PIHandlerImpl.REPLY_MESSAGE_TO_PI_REPLY_STATUS[replyMessage.getReplyStatus()];
            if (replyStatus == 3 || replyStatus == 4) {
                peekServerRequestInfoImplStack.setForwardRequest(replyMessage.getIOR());
            }
            Exception exception = peekServerRequestInfoImplStack.getException();
            if (!peekServerRequestInfoImplStack.isDynamic() && replyStatus == 2) {
                peekServerRequestInfoImplStack.setException(this.omgWrapper.unknownUserException(CompletionStatus.COMPLETED_MAYBE));
            }
            peekServerRequestInfoImplStack.setReplyStatus((short)replyStatus);
            this.interceptorInvoker.invokeServerInterceptorEndingPoint(peekServerRequestInfoImplStack);
            int replyStatus2 = peekServerRequestInfoImplStack.getReplyStatus();
            Exception exception2 = peekServerRequestInfoImplStack.getException();
            if (replyStatus2 == 1 && exception2 != exception) {
                throw (SystemException)exception2;
            }
            if (replyStatus2 == 3) {
                if (replyStatus != 3) {
                    throw new ForwardException(this.orb, peekServerRequestInfoImplStack.getForwardRequestIOR());
                }
                if (peekServerRequestInfoImplStack.isForwardRequestRaisedInEnding()) {
                    replyMessage.setIOR(peekServerRequestInfoImplStack.getForwardRequestIOR());
                }
            }
        }
        return;
    }
    
    public void setServerPIInfo(Exception exception) {
        if (!this.hasServerInterceptors) {
            return;
        }
        this.peekServerRequestInfoImplStack().setException(exception);
        return;
    }
    
    public void setServerPIInfo(NVList dsiArguments) {
        if (!this.hasServerInterceptors) {
            return;
        }
        this.peekServerRequestInfoImplStack().setDSIArguments(dsiArguments);
        return;
    }
    
    public void setServerPIExceptionInfo(Any dsiException) {
        if (!this.hasServerInterceptors) {
            return;
        }
        this.peekServerRequestInfoImplStack().setDSIException(dsiException);
        return;
    }
    
    public void setServerPIInfo(Any dsiResult) {
        if (!this.hasServerInterceptors) {
            return;
        }
        this.peekServerRequestInfoImplStack().setDSIResult(dsiResult);
        return;
    }
    
    public void initializeServerPIInfo(CorbaMessageMediator corbaMessageMediator, ObjectAdapter objectAdapter, byte[] array, ObjectKeyTemplate objectKeyTemplate) {
        if (!this.hasServerInterceptors) {
            return;
        }
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalServerRequestInfoStack.get();
        ServerRequestInfoImpl item = new ServerRequestInfoImpl(this.orb);
        requestInfoStack.push(item);
        this.printPush();
        corbaMessageMediator.setExecutePIInResponseConstructor(true);
        item.setInfo(corbaMessageMediator, objectAdapter, array, objectKeyTemplate);
        return;
    }
    
    public void setServerPIInfo(Object o, String s) {
        if (!this.hasServerInterceptors) {
            return;
        }
        this.peekServerRequestInfoImplStack().setInfo(o, s);
        return;
    }
    
    public void cleanupServerPIRequest() {
        if (!this.hasServerInterceptors) {
            return;
        }
        ((PIHandlerImpl.RequestInfoStack)this.threadLocalServerRequestInfoStack.get()).pop();
        this.printPop();
        return;
    }
    
    private void serverPIHandleExceptions(ServerRequestInfoImpl serverRequestInfoImpl) {
        int endingPointCall = serverRequestInfoImpl.getEndingPointCall();
        if (endingPointCall == 1) {
            throw (SystemException)serverRequestInfoImpl.getException();
        }
        if (endingPointCall == 2 && serverRequestInfoImpl.getForwardRequestException() != null) {
            throw new ForwardException(this.orb, serverRequestInfoImpl.getForwardRequestIOR());
        }
        return;
    }
    
    private int convertPIReplyStatusToReplyMessage(short n) {
        int n2 = 0;
        int i = 0;
        while (i < PIHandlerImpl.REPLY_MESSAGE_TO_PI_REPLY_STATUS.length) {
            if (PIHandlerImpl.REPLY_MESSAGE_TO_PI_REPLY_STATUS[i] == n) {
                n2 = i;
                break;
            }
            ++i;
        }
        return n2;
    }
    
    private ClientRequestInfoImpl peekClientRequestInfoImplStack() {
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get();
        if (!requestInfoStack.empty()) {
            return (ClientRequestInfoImpl)requestInfoStack.peek();
        }
        throw this.wrapper.clientInfoStackNull();
    }
    
    private ServerRequestInfoImpl peekServerRequestInfoImplStack() {
        PIHandlerImpl.RequestInfoStack requestInfoStack = (PIHandlerImpl.RequestInfoStack)this.threadLocalServerRequestInfoStack.get();
        if (!requestInfoStack.empty()) {
            return (ServerRequestInfoImpl)requestInfoStack.peek();
        }
        throw this.wrapper.serverInfoStackNull();
    }
    
    private boolean isClientPIEnabledForThisThread() {
        return ((PIHandlerImpl.RequestInfoStack)this.threadLocalClientRequestInfoStack.get()).disableCount == 0;
    }
    
    private void preInitORBInitializers(ORBInitInfoImpl orbInitInfoImpl) {
        orbInitInfoImpl.setStage(0);
        for (int i = 0; i < this.orb.getORBData().getORBInitializers().length; ++i) {
            ORBInitializer orbInitializer = this.orb.getORBData().getORBInitializers()[i];
            if (orbInitializer != null) {
                try {
                    orbInitializer.pre_init(orbInitInfoImpl);
                }
                catch (Exception ex) {}
            }
        }
    }
    
    private void postInitORBInitializers(ORBInitInfoImpl orbInitInfoImpl) {
        orbInitInfoImpl.setStage(1);
        for (int i = 0; i < this.orb.getORBData().getORBInitializers().length; ++i) {
            ORBInitializer orbInitializer = this.orb.getORBData().getORBInitializers()[i];
            if (orbInitializer != null) {
                try {
                    orbInitializer.post_init(orbInitInfoImpl);
                }
                catch (Exception ex) {}
            }
        }
    }
    
    private ORBInitInfoImpl createORBInitInfo() {
        return new ORBInitInfoImpl(this.orb, this.arguments, this.orb.getORBData().getORBId(), this.codecFactory);
    }
    
    public void register_interceptor(Interceptor interceptor, int value) throws DuplicateName {
        if (value >= 3 || value < 0) {
            throw this.wrapper.typeOutOfRange(new Integer(value));
        }
        if (interceptor.name() == null) {
            throw this.wrapper.nameNull();
        }
        this.interceptorList.register_interceptor(interceptor, value);
        return;
    }
    
    public Current getPICurrent() {
        return this.current;
    }
    
    private void nullParam() throws BAD_PARAM {
        throw this.orbutilWrapper.nullParam();
    }
    
    public Policy create_policy(int i, Any any) throws PolicyError {
        if (any == null) {
            this.nullParam();
        }
        if (this.policyFactoryTable == null) {
            throw new PolicyError("There is no PolicyFactory Registered for type " + i, (short)0);
        }
        PolicyFactory policyFactory = (PolicyFactory)this.policyFactoryTable.get(new Integer(i));
        if (policyFactory == null) {
            throw new PolicyError(" Could Not Find PolicyFactory for the Type " + i, (short)0);
        }
        return policyFactory.create_policy(i, any);
    }
    
    public void registerPolicyFactory(int n, PolicyFactory value) {
        if (this.policyFactoryTable == null) {
            this.policyFactoryTable = new HashMap();
        }
        Integer n2 = new Integer(n);
        if (this.policyFactoryTable.get(n2) == null) {
            this.policyFactoryTable.put(n2, value);
            return;
        }
        throw this.omgWrapper.policyFactoryRegFailed(new Integer(n));
    }
    
    public synchronized int allocateServerRequestId() {
        return this.serverRequestIdCounter++;
    }
    
    static {
        REPLY_MESSAGE_TO_PI_REPLY_STATUS = new short[] { 0, 2, 1, 3, 3, 4 };
    }
    
    private final class RequestInfoStack extends Stack
    {
        public int disableCount;
        
        private RequestInfoStack() {
            this.disableCount = 0;
        }
    }
}
